from django.contrib.auth import authenticate, get_user_model
from rest_framework import exceptions, serializers
from django.contrib.auth import get_user_model
from django.contrib.auth.models import update_last_login
from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions, serializers
from rest_framework_simplejwt.settings import api_settings
from rest_framework_simplejwt.tokens import RefreshToken, SlidingToken, UntypedToken
from rest_framework_simplejwt.serializers import PasswordField
from rest_framework_jwt.settings import api_settings as jwt_settings

from .token_manager import CodeTokenManager
from .utils import check_user_validity

jwt_encode_handler = jwt_settings.JWT_ENCODE_HANDLER
jwt_payload_handler = jwt_settings.JWT_PAYLOAD_HANDLER

from rest_framework import serializers


class JwtSerializer(serializers.Serializer):
    @property
    def object(self):
        return self.validated_data


class Jwt2faSerializer(JwtSerializer):
    token_manager_class = CodeTokenManager

    def __init__(self, *args, **kwargs):
        super(Jwt2faSerializer, self).__init__(*args, **kwargs)
        self.token_manager = self.token_manager_class()

    def validate(self, attrs):
        validated_attrs = super(Jwt2faSerializer, self).validate(attrs)
        user = self._authenticate(validated_attrs)
        return {
            'token': self._create_token(user),
        }


class CodeTokenSerializer(Jwt2faSerializer):
    username = serializers.CharField(required=True)
    password = PasswordField(write_only=True, required=True)

    def _authenticate(self, attrs):
        credentials = {
            'username': attrs.get('username'),
            'password': attrs.get('password'),
        }
        user = authenticate(**credentials)
        if not user:
            raise exceptions.AuthenticationFailed()
        check_user_validity(user)
        return user

    def _create_token(self, user):
        return self.token_manager.create_code_token(user)


class AuthTokenSerializer(Jwt2faSerializer):
    code_token = serializers.CharField(required=True)
    code = PasswordField(write_only=True, required=True)

    def _authenticate(self, attrs):
        code_token = attrs.get('code_token')
        code = attrs.get('code')
        username = self._check_code_token_and_code(code_token, code)
        user = self._get_user(username)
        return user

    def _check_code_token_and_code(self, code_token, code):
        return self.token_manager.check_code_token_and_code(code_token, code)

    def _get_user(self, username):
        user_model = get_user_model()
        try:
            user = user_model.objects.get_by_natural_key(username)
        except user_model.DoesNotExist:
            raise exceptions.AuthenticationFailed()
        check_user_validity(user)
        return user

    def _create_token(self, user):
        payload = jwt_payload_handler(user)
        return jwt_encode_handler(payload)


class TokenObtainSerializer(serializers.Serializer):
    token_manager_class = CodeTokenManager

    default_error_messages = {
        "no_active_account": _("No active account found with the given credentials")
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["code_token"] = serializers.CharField()
        self.fields["code"] = serializers.CharField()

    def _get_user(self, username):
        user_model = get_user_model()
        try:
            user = user_model.objects.get_by_natural_key(username)
        except user_model.DoesNotExist:
            raise exceptions.AuthenticationFailed()
        check_user_validity(user)
        return user

    def _authenticate(self, attrs):
        code_token = attrs.get("code_token")
        code = attrs.get("code")
        username = self.token_manager_class().check_code_token_and_code(code_token, code)
        user = self._get_user(username)
        return user

    def validate(self, attrs):
        authenticate_kwargs = {
            "code_token": attrs["code_token"],
            "code": attrs["code"],
        }
        try:
            authenticate_kwargs["request"] = self.context["request"]
        except KeyError:
            pass

        self.user = self._authenticate(authenticate_kwargs)

        if not default_user_authentication_rule(self.user):
            raise exceptions.AuthenticationFailed(
                self.error_messages["no_active_account"],
                "no_active_account",
            )

        return {}

    @classmethod
    def get_token(cls, user):
        raise NotImplementedError(
            "Must implement `get_token` method for `TokenObtainSerializer` subclasses"
        )


class TokenObtainPairSerializer(TokenObtainSerializer):
    @classmethod
    def get_token(cls, user):
        return RefreshToken.for_user(user)

    def validate(self, attrs):
        data = super().validate(attrs)

        refresh = self.get_token(self.user)

        data["refresh"] = str(refresh)
        data["access"] = str(refresh.access_token)

        if api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, self.user)

        return data


class TokenObtainSlidingSerializer(TokenObtainSerializer):
    @classmethod
    def get_token(cls, user):
        return SlidingToken.for_user(user)

    def validate(self, attrs):
        data = super().validate(attrs)

        token = self.get_token(self.user)

        data["token"] = str(token)

        if api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, self.user)

        return data


def default_user_authentication_rule(user):
    # Prior to Django 1.10, inactive users could be authenticated with the
    # default `ModelBackend`.  As of Django 1.10, the `ModelBackend`
    # prevents inactive users from authenticating.  App designers can still
    # allow inactive users to authenticate by opting for the new
    # `AllowAllUsersModelBackend`.  However, we explicitly prevent inactive
    # users from authenticating to enforce a reasonable policy and provide
    # sensible backwards compatibility with older Django versions.
    return user is not None and user.is_active


from rest_framework_simplejwt import views as jwt_views

from . import serializers
from .throttling import AuthTokenThrottler, CodeTokenThrottler


class ObtainCodeToken(jwt_views.TokenObtainPairView):
    serializer_class = serializers.CodeTokenSerializer
    throttle_classes = [CodeTokenThrottler]


class ObtainAuthToken(jwt_views.TokenObtainPairView):
    serializer_class = serializers.TokenObtainPairSerializer
    throttle_classes = [AuthTokenThrottler]


class RefreshAuthToken(jwt_views.TokenRefreshView):
    pass


class VerifyAuthToken(jwt_views.TokenVerifyView):
    pass


obtain_code_token = ObtainCodeToken.as_view()
obtain_auth_token = ObtainAuthToken.as_view()
refresh_auth_token = RefreshAuthToken.as_view()
verify_auth_token = VerifyAuthToken.as_view()



from django.conf import settings
from django.core.mail import send_mail
from django.utils.translation import ugettext as _
import re

from .settings import api_settings


class CodeSendingFailed(Exception):
    pass


def send_verification_code(user, code):
    sender = api_settings.CODE_SENDER
    return sender(user, code)


def send_verification_code_via_email(user, code):
    user_email_address = getattr(user, 'first_name', None)
    regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
    if not re.fullmatch(regex, user_email_address):
        user_email_address = getattr(user, 'email', None)
        if not re.fullmatch(regex, user_email_address):
            raise CodeSendingFailed(_("No e-mail address known"))

    subject_template = _(
        api_settings.EMAIL_SENDER_SUBJECT_OVERRIDE or
        _("{code}: Your verification code"))
    body_template = (
        api_settings.EMAIL_SENDER_BODY_OVERRIDE or
        _("{code} is the verification code needed for the login."))

    messages_sent = send_mail(
        subject=subject_template.format(code=code),
        message=body_template.format(code=code),
        from_email=settings.DEFAULT_FROM_EMAIL,
        recipient_list=[user_email_address],
        fail_silently=True)

    if not messages_sent:
        raise CodeSendingFailed(_("Unable to send e-mail"))


import pytest
from django.contrib.auth.backends import ModelBackend
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from .factories import get_user
from .utils import (
    check_auth_token, check_code_token, get_api_client,
    get_verification_code_from_mailbox)


@pytest.mark.django_db
def test_get_code_token_success():
    token = get_code_token()
    check_code_token(token)


def get_code_token():
    get_user(username='testuser', password='a42')
    client = get_api_client()
    result = client.post(
        reverse('get-code'),
        data={'username': 'testuser', 'password': 'a42'})
    assert sorted(result.data.keys()) == ['token']
    assert result.status_code == status.HTTP_200_OK
    return result.data['token']


def test_code_token_missing_fields():
    client = get_api_client()
    # Post without username field
    result = client.post(reverse('get-code'), data={'password': 'abc'})
    assert sorted(result.data.keys()) == ['username']
    assert result.status_code == status.HTTP_400_BAD_REQUEST
    assert result.data['username'] == ['This field is required.']


@pytest.mark.django_db
def test_code_token_invalid_password():
    get_user(username='testuser', password='a42')
    client = get_api_client()
    result = client.post(
        reverse('get-code'),
        data={'username': 'testuser', 'password': 'wrong'})
    assert result.data == {'detail': 'Incorrect authentication credentials.'}
    assert result.status_code == status.HTTP_403_FORBIDDEN


class InactiveAllowingAuthBackend(ModelBackend):
    def user_can_authenticate(self, user):
        return True


@override_settings(AUTHENTICATION_BACKENDS=[
    __name__ + '.InactiveAllowingAuthBackend'])
@pytest.mark.django_db
def test_code_token_inactive_user():
    user = get_user(username='testuser', password='a42')
    user.is_active = False
    user.save()
    client = get_api_client()
    result = client.post(
        reverse('get-code'),
        data={'username': 'testuser', 'password': 'a42'})
    assert result.data == {
        'detail': 'You do not have permission to perform this action.'}
    assert result.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.django_db
def test_auth_token_success():
    code_token = get_code_token()
    code = get_verification_code_from_mailbox()
    client = get_api_client()
    result = client.post(
        reverse('auth'),
        data={'code_token': code_token, 'code': code})
    assert sorted(result.data.keys()) == ['token']
    assert result.status_code == status.HTTP_200_OK
    token = result.data['token']
    check_auth_token(token)


@pytest.mark.django_db
def test_auth_token_invalid_code():
    code_token = get_code_token()
    correct_code = get_verification_code_from_mailbox()
    code = '1234567' if correct_code != '1234567' else '7654321'
    client = get_api_client()
    result = client.post(
        reverse('auth'),
        data={'code_token': code_token, 'code': code})
    assert result.data == {'detail': 'Incorrect authentication credentials.'}
    assert result.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.django_db
def test_auth_token_removed_user():
    code_token = get_code_token()
    code = get_verification_code_from_mailbox()
    user = get_user()
    user.delete()
    client = get_api_client()
    result = client.post(
        reverse('auth'),
        data={'code_token': code_token, 'code': code})
    assert result.data == {'detail': 'Incorrect authentication credentials.'}
    assert result.status_code == status.HTTP_403_FORBIDDEN



from __future__ import unicode_literals

import datetime
import time

import pytest
from django.core import mail
from mock import patch
from rest_framework import exceptions, status

from drf_jwt_2fa.exceptions import VerificationCodeSendingFailed
from drf_jwt_2fa.token_manager import CodeTokenManager

from .factories import get_user
from .utils import (
    OverrideJwt2faSettings, check_code_token, decode_jwt_part, encode_jwt_part,
    get_verification_code_from_mailbox)


@pytest.mark.django_db
def test_create_code_token():
    manager = CodeTokenManager()

    mail_outbox_size_before = len(mail.outbox)

    token = manager.create_code_token(get_user())

    # Check sent mails
    assert len(mail.outbox) == mail_outbox_size_before + 1
    assert mail.outbox[-1].subject.endswith(': Your verification code')
    assert mail.outbox[-1].from_email == 'webmaster@localhost'
    assert mail.outbox[-1].to == ['testuser@localhost']

    # Check the generated token
    check_code_token(token)


@pytest.mark.django_db
def test_create_code_token_with_no_email():
    manager = CodeTokenManager()

    mail_outbox_size_before = len(mail.outbox)

    with pytest.raises(VerificationCodeSendingFailed) as exc_info:
        manager.create_code_token(get_user(username='no-email', email=''))
    assert str(exc_info.value) == (
        'Verification code sending failed: No e-mail address known')
    assert exc_info.value.status_code == status.HTTP_501_NOT_IMPLEMENTED

    # Check sent mails
    assert len(mail.outbox) == mail_outbox_size_before


@pytest.mark.django_db
@patch('drf_jwt_2fa.sending.send_mail', return_value=0)
def test_create_code_token_with_email_send_error(mocked_send_mail):
    manager = CodeTokenManager()

    with pytest.raises(VerificationCodeSendingFailed) as exc_info:
        manager.create_code_token(get_user())
    assert str(exc_info.value) == (
        'Verification code sending failed: Unable to send e-mail')
    assert exc_info.value.status_code == status.HTTP_501_NOT_IMPLEMENTED

    assert mocked_send_mail.called_once()


@pytest.mark.django_db
def test_check_code_token_and_code_success():
    manager = CodeTokenManager()
    token = manager.create_code_token(get_user())
    code = get_verification_code_from_mailbox()
    username = manager.check_code_token_and_code(token, code)
    assert username == 'testuser'


@pytest.mark.django_db
def test_check_code_token_and_code_with_invalid_token():
    manager = CodeTokenManager()
    token = manager.create_code_token(get_user())
    code = get_verification_code_from_mailbox()
    (header, payload_before, signature) = token.split('.')
    payload = decode_jwt_part(payload_before)
    payload['usr'] = 'somebody-else'
    new_token = header + '.' + encode_jwt_part(payload) + '.' + signature
    check_code_token(new_token, username='somebody-else', verify=False)
    with pytest.raises(exceptions.AuthenticationFailed) as exc_info:
        manager.check_code_token_and_code(new_token, code)
    assert str(exc_info.value) == 'Incorrect authentication credentials.'
    assert exc_info.value.status_code == status.HTTP_401_UNAUTHORIZED


@pytest.mark.django_db
def test_check_code_token_and_code_with_invalid_code():
    manager = CodeTokenManager()
    token = manager.create_code_token(get_user())
    correct_code = get_verification_code_from_mailbox()
    assert len(correct_code) == 7
    invalid_code = '1234567' if correct_code != '1234567' else '7654321'
    with pytest.raises(exceptions.AuthenticationFailed) as exc_info:
        manager.check_code_token_and_code(token, invalid_code)
    assert str(exc_info.value) == 'Incorrect authentication credentials.'
    assert exc_info.value.status_code == status.HTTP_401_UNAUTHORIZED


@pytest.mark.django_db
@OverrideJwt2faSettings({
    'CODE_EXPIRATION_TIME': datetime.timedelta(seconds=-1),
})
def test_check_code_token_and_code_with_expired_token():
    manager = CodeTokenManager()
    token = manager.create_code_token(get_user())
    assert decode_jwt_part(token.split('.')[1])['exp'] < time.time()
    code = get_verification_code_from_mailbox()
    with pytest.raises(exceptions.PermissionDenied) as exc_info:
        manager.check_code_token_and_code(token, code)
    assert str(exc_info.value) == 'Signature has expired.'
    assert exc_info.value.status_code == status.HTTP_403_FORBIDDEN



from django.conf.urls import url

from . import views

urlpatterns = [
    url('^get-code/', views.obtain_code_token, name='get-code'),
    url('^auth/', views.obtain_auth_token, name='auth'),
    url('^refresh/', views.refresh_auth_token, name='refresh'),
    url('^verify/', views.verify_auth_token, name='verify'),
]



from django.contrib.auth.models import User


def get_user(username='testuser', password='a42', email='testuser@localhost'):
    user = User.objects.get_or_create(username=username)[0]
    user.set_password(password)
    user.email = email
    user.is_active = True
    user.save()
    return user
